# Configuration for networking init script being run during
# the boot sequence

# Set to 'no' to skip interfaces configuration on boot
CONFIGURE_INTERFACES={{networking}}

# Don't configure these interfaces. Shell wildcards supported/
#EXCLUDE_INTERFACES=

# Set to 'yes' to enable additional verbosity
#VERBOSE=no
