# Type Path                    Mode UID  GID     Age Argument
d      /var/run/salt           0750 root salt
d      /var/run/salt/master    0750 salt salt
d      /var/run/salt/minion    0750 root root
